/*
	Game Categories v 12.0
	Copyright (C) 2009, Bubbletune

	communication.c: Exports for the user module
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <pspsdk.h>
#include <pspkernel.h>
#include <string.h>
#include <systemctrl.h>
#include <systemctrl_se.h>
#include <psputilsforkernel.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <game_categories.h>

u32 gcLibKernelFindProc(const char* szMod, const char* szLib, u32 nid)
{
	return sctrlHENFindFunction(szMod, szLib, nid);
}

int gcLibKernelGetDevkitID(char *modname)
{
	int i;

	u32 devkit;
	u32 *devkit_value = (u32 *)gcLibKernelFindProc(modname, NULL, 0x11B97506);

	if (devkit_value)
	{
		devkit = *devkit_value;
	}
	else
	{
		devkit = sceKernelDevkitVersion();
	}

	static u32 devkits[] =
	{
		0x05000010, // release_500
		0x05050010, // release_550
	};

	for (i = sizeof(devkits)/sizeof(u32); i > -1; i--)
	{
		if (devkit >= devkits[i])
		{
			return i;
		}
	}
	return -1;
}

#define MASKBITS   0x3F
#define MASKBYTE   0x80
#define MASK2BYTES 0xC0
#define MASK3BYTES 0xE0
#define MASK4BYTES 0xF0
#define MASK5BYTES 0xF8
#define MASK6BYTES 0xFC

int gcLibKernelUTF8ToUnicode(wchar_t *dest, char *src)
{
	int i, x;
	unsigned char *usrc = (unsigned char *)src;

	for(i = 0, x = 0; usrc[i];)
	{
		wchar_t ch;

		if((usrc[i] & MASK3BYTES) == MASK3BYTES)
		{
			ch = ((usrc[i] & 0x0F) << 12) | ((usrc[i+1] & MASKBITS) << 6) | (usrc[i+2] & MASKBITS);
			i += 3;
		}
		else if((usrc[i] & MASK2BYTES) == MASK2BYTES)
		{
			ch = ((usrc[i] & 0x1F) << 6) | (usrc[i+1] & MASKBITS);
			i += 2;
		}
		else/* if(usrc[i] < MASKBYTE)*/
		{
			ch = usrc[i];
			i += 1;
		}
		dest[x++] = ch;
	}
	dest[x++] = '\0';
	return x;
}

SceModule2 *gcLibKernelFindModuleByName(char *modname)
{
	return sceKernelFindModuleByName(modname);
}

u32 gcLibKernelPatchExport(char *modname, char *importlib, u32 nid, void *newone) 
{
	SceModule2 *mod = sceKernelFindModuleByName(modname);

	if (mod)
	{
		int i = 0, u;
		int entry_size = mod->ent_size;
		int entry_start = (int)mod->ent_top;

		while (i < entry_size)
		{
			SceLibraryEntryTable *entry = (SceLibraryEntryTable *)(entry_start + i);

			if (entry->libname && (strcmp(entry->libname, importlib) == 0))
			{
				u32 *table = entry->entrytable;
				int total = entry->stubcount + entry->vstubcount;

				if (total > 0)
				{ 
					for (u = 0; u < total; u++)
					{ 
						if (table[u] == nid)
						{
							table[u + total] = (u32)newone;
							return 1;
						}
					}
				} 	
			}
			i += (entry->len * 4);
		}
	}
	return 0;
}

