/*
	Game Categories v 12.0
	Copyright (C) 2009, Bubbletune

	main.c: User memory main code
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <pspsdk.h>
#include <pspkernel.h>

#include <systemctrl.h>
#include <systemctrl_se.h>
#include <game_categories.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

PSP_MODULE_INFO("XMBattery", 0x0007, 0, 0);
PSP_NO_CREATE_MAIN_THREAD();

u32 sysconf_page = 0;
char user_buffer[256];
STMOD_HANDLER previous = NULL;
SceSysconfItem *batop = NULL;
u32 vshmain_addr;

char *batop_text[] = {
	"Pandora",
	"Normal",
	"Autoboot"
};

int XMBSetMode(int mode);
int XMBGetMode(void);

void (* pspKernelIcacheClearAll)(void);

void ClearCaches()
{
	sceKernelDcacheWritebackAll();
	pspKernelIcacheClearAll();
}

u32 ReleaseContext(u32 id)
{
	if (id && id != 1)
	{
		u32 *buffer = (u32 *)id;

		if (gcLibKernelFindModuleByName("sysconf_plugin_module"))
		{
			SceRcoEntry *src = (SceRcoEntry *)buffer[0];
			SceRcoEntry *plane = (SceRcoEntry *)(((u8 *)src)+src->first_child);
			SceRcoEntry *mlist = (SceRcoEntry *)(((u8 *)plane)+plane->first_child);
			u32 *mlist_param = (u32 *)(((u8 *)mlist)+mlist->param);

			mlist->first_child = buffer[1];
			mlist->child_count = buffer[2];
			mlist_param[16] = buffer[3];
			mlist_param[18] = buffer[4];
		}
		paf_free(buffer);
		sceKernelDcacheWritebackAll();
	}
	return 0;
}

int vshGetRegistryValuePatched(u32 *option, char *name, void *arg2, int size, int *value)
{
	if (name)
	{
		sysconf_page = ReleaseContext(sysconf_page);
		if(strcmp(name,"bat_ops") == 0)
		{
			sysconf_page = 1;
			*value = XMBGetMode();
			return 0;
		}
	}
	return vshGetRegistryValue(option, name, arg2, size, value);
}

int vshSetRegistryValuePatched(u32 *option, char *name, int size, int *value)
{
	if (name)
	{
		if(strcmp(name,"bat_ops") == 0)
		{
			XMBSetMode(*value);
			return 0;
		}
	}
	return vshSetRegistryValue(option, name, size, value);
}

u32 HijackContext(void *resource, SceRcoEntry *src, char **options, int n)
{
	int i;
	u32 *buffer;
	SceRcoEntry *plane, *mlist, *base, *item;
	u32 *page_param, *mlist_param, *item_param;

	page_param = (u32 *)(((u8 *)src)+src->param);
	plane = (SceRcoEntry *)(((u8 *)src)+src->first_child);
	mlist = (SceRcoEntry *)(((u8 *)plane)+plane->first_child);
	mlist_param = (u32 *)(((u8 *)mlist)+mlist->param);
	base = (SceRcoEntry *)(((u8 *)mlist)+mlist->first_child);

	buffer = paf_malloc(20+(base->next_entry*n));
	item = (SceRcoEntry *)(((u8 *)buffer)+20);
	item_param = (u32 *)(((u8 *)item)+base->param);

	buffer[0] = (u32)src;
	buffer[1] = mlist->first_child;
	buffer[2] = mlist->child_count;
	buffer[3] = mlist_param[16];
	buffer[4] = mlist_param[18];

	mlist->first_child = (u32)(((u8 *)item)-((u8 *)mlist));
	mlist->child_count = n;
	mlist_param[16] = (n+1)+(n%2);
	mlist_param[18] = (n/2)+(n%2);

	for (i = 0; i < n; i++)
	{
		paf_memcpy(item, base, base->next_entry);

		item_param[0] = 0xDEAD;
		item_param[1] = (u32)options[i];

		if (i != 0)
		{
			item->prev_entry = item->next_entry;
		}

		if (i == n-1)
		{
			item->next_entry = 0;
		}

		item = (SceRcoEntry *)(((u8 *)item)+base->next_entry);
		item_param = (u32 *)(((u8 *)item)+base->param);
	}

	sceKernelDcacheWritebackAll();
	return (u32)buffer;
}

int scePafGetPageStringPatched(void *resource, u32 *data, int *arg2, char **string, int *temp0)
{
	if (data[0] == 0xDEAD)
	{
		gcLibKernelUTF8ToUnicode((wchar_t*)user_buffer, (char *)data[1]);
		*(wchar_t **)string = (wchar_t*)user_buffer;
		return 0;
	}
	return scePafGetPageString(resource, data, arg2, string, temp0);
}

int scePafGetPageChildPatched(void *resource, SceRcoEntry *parent, char *name, SceRcoEntry **child)
{
	int res = scePafGetPageChild(resource, parent, name, child);

	if (name)
	{
		if (strcmp(name, "page_psp_config_umd_autoboot") == 0)
		{
			if (sysconf_page == 1)
			{
				sysconf_page = HijackContext(resource, *child, batop_text, sizeof(batop_text)/sizeof(char*));
			}
		}
	}
	return res;
}

wchar_t* scePafGetTextPatched(void *arg, char *name)
{
	if (name)
	{
		if (strcmp(name,"bat_ops_txt") == 0)
		{
			gcLibKernelUTF8ToUnicode((wchar_t*)user_buffer, "Battery Mode");
			return (wchar_t*)user_buffer;
		}
	}
	return scePafGetText(arg, name);
}

void PatchPafModule(u32 text_addr)
{
	int i;
	int id = gcLibKernelGetDevkitID("scePaf_Module");

	if (id >= 0)
	{
		gcLibKernelPatchExport("scePaf_Module", "scePaf", 0x970DC20D, scePafGetTextPatched);

		static addr_based_patch paf_patches[] =
		{
			{ { 0x372F0, 0x372F0, }, NULL, scePafGetPageChildPatched },
			{ { 0x66DF4, 0x66DF4, }, NULL, scePafGetPageChildPatched },
			{ { 0x3A434, 0x3A434, }, NULL, scePafGetPageStringPatched },
			{ { 0x67090, 0x067090, }, NULL, scePafGetPageStringPatched },
		};

		for (i = 0; i < sizeof(paf_patches)/sizeof(addr_based_patch); i++)
		{
			if (paf_patches[i].original)
				MAKE_JUMP((u32)paf_patches[i].original, U_EXTRACT_CALL(text_addr+paf_patches[i].patch[id]));

			MAKE_CALL(text_addr+paf_patches[i].patch[id], paf_patches[i].patched);
		}
		ClearCaches();
	}
}

SceSysconfItem *GetSysconfItemPatched(void *arg0, void *arg1)
{
	SceSysconfItem *item = GetSysconfItem(arg0, arg1);

	sysconf_page = ReleaseContext(sysconf_page);

	if (strcmp(item->text, "bat_ops_txt") == 0)
	{
		sysconf_page = 1;
	}
	return item;
}

void AddSysconfItemPatched(u32 *option, SceSysconfItem **item)
{
	AddSysconfItem(option, item);

	if (!batop)
		batop = (void*)paf_malloc(sizeof(SceSysconfItem));

	paf_memcpy(batop, *item, sizeof(SceSysconfItem));
	batop->id = 3;
	batop->text = "bat_ops_txt";
	batop->regkey = "bat_ops";
	batop->page = "page_psp_config_umd_autoboot";

	option[2] = 1;
	AddSysconfItem(option, &batop);
}

int ExecuteActionPatched(int action, int action_arg)
{
	if (!gcLibKernelFindModuleByName("sysconf_plugin_module"))
	{
		batop = NULL;
	}
	return ExecuteAction(action, action_arg);
}

void PatchVshmain(u32 text_addr)
{
	int i;
	int id = gcLibKernelGetDevkitID("vsh_module");
	vshmain_addr = text_addr;
	if (id >= 0)
	{
		static nid_based_patch vshmain_patches[] =
		{
			{ "vsh_module", "vsh", 0x55DD305F, vshGetRegistryValuePatched },
			{ "vsh_module", "vsh", 0x0A9043D4, vshSetRegistryValuePatched },
		};

		for (i = 0; i < sizeof(vshmain_patches)/sizeof(nid_based_patch); i++)
		{
			if (!gcLibKernelPatchExport(vshmain_patches[i].modname, vshmain_patches[i].libname, vshmain_patches[i].nid, vshmain_patches[i].func))
			{
				return;
			}
		}

		static addr_based_patch vshmain_patches_2[] =
		{
			{ { 0x12650, 0x13270, }, ExecuteAction, ExecuteActionPatched },
			{ { 0x266A8, 0x2767C, }, NULL, ExecuteActionPatched },
		};

		for (i = 0; i < sizeof(vshmain_patches_2)/sizeof(addr_based_patch); i++)
		{
			if (vshmain_patches_2[i].original)
				MAKE_JUMP((u32)vshmain_patches_2[i].original, U_EXTRACT_CALL(text_addr+vshmain_patches_2[i].patch[id]));
				
			MAKE_CALL(text_addr+vshmain_patches_2[i].patch[id], vshmain_patches_2[i].patched);
		}

		ClearCaches();
	}
}

void PatchSysconf(u32 text_addr)
{
	int i;
	int id = gcLibKernelGetDevkitID("sysconf_plugin_module");

	if (id >= 0)
	{
		static addr_based_patch sysconf_patches[] =
		{
			{ { 0x19610, 0x1943C, }, AddSysconfItem, AddSysconfItemPatched },
			{ { 0x0241C, 0x02414, }, GetSysconfItem, GetSysconfItemPatched },
		};

		for (i = 0; i < sizeof(sysconf_patches)/sizeof(addr_based_patch); i++)
		{
			MAKE_JUMP((u32)sysconf_patches[i].original, U_EXTRACT_CALL(text_addr+sysconf_patches[i].patch[id]));
			MAKE_CALL(text_addr+sysconf_patches[i].patch[id], sysconf_patches[i].patched);
		}
		ClearCaches();
	}
}

int OnModuleStart(SceModule2 *mod)
{
	if (strcmp(mod->modname, "scePaf_Module") == 0)
	{
		PatchPafModule(mod->text_addr);
	}
	else if (strcmp(mod->modname, "vsh_module") == 0)
	{
		PatchVshmain(mod->text_addr);
	}
	else if (strcmp(mod->modname, "sysconf_plugin_module") == 0)
	{
		PatchSysconf(mod->text_addr);
	}

	if (!previous)
		return 0;

	return previous(mod);
}

int module_start(SceSize args, void *argp)
{
	pspKernelIcacheClearAll = *(void **)argp;
	previous = sctrlHENSetStartModuleHandler(OnModuleStart);
	return 0;
}
